/*[]=====================================================================[]*/
/*[]     LeadTools Run Time Library - Version 3.6                        []*/
/*[]                                                                     []*/
/*[]                                                                     []*/
/*[]     Copyright (c) 1990, 1993 by LEAD Technologies, Inc.             []*/
/*[]     All Rights Reserved.                                            []*/
/*[]=====================================================================[]*/

/*---(L_RedirectIO)---------------------------------------------------------
  We have made the assumption that the user has the knowledge of programing
  in C and Windows.


  This example Will show how to use the power of redirect I/O feature 
  in our DLL. The concept behind the redirecting process is to allow 
  the user to deal with image data ANY WHERE, (ANY WHERE means any where)
  either it is in disk, memory, I/O com port, ... Use your imagination !!!

  To demonstrate that we load a compressed image file passed as a parameter
  to this program, into a located memory. Then we use L_DecompressBitmap
  to decompress the image and display it into a window.

  Before calling L_DecompressBitmap we use L_RedirectIO to redirect all
  I/O operations to our I/O functions. Our I/O functions will read and seek
  into the allocated memory rather than a file. Because we need to access 
  a global variables, we exported these function so the DS can be restored.

  It is important to export your I/O functions by using the word _export
  in the function definition, and Create an instance of these functions.
  The created instance will be used as the address for that function.

   This is a generic example which means it will show how to solve your
   problem by extending these file I/O functions, we don't cover all 
   the cases rather than we don't assume it may handle any writing 
   operation.

  Usage: DECOMBIT <filename>

  --------------------------------------------------------------------------*/

#include <windows.h>		    /* required for all Windows applications */
#include <mmsystem.h>		 /* required for all multi media calls    */
#include <stdlib.h>         /* Standard C header files.              */
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>

#define FOR_DLL            /* Tell L_Bitmap.h to include only DLL specific definitions */
#include "l_bitmap.h"      /* Lead Tools main header file */
#include "l_error.h"       /* Lead Tools error definition header file */
#include "redirect.h"      /* application specific header file */

DATA     pData;                 /* Pointer to DATA structure */
HWND     hWndFrame;             /* Main window handle.       */
HANDLE   hInst;			        /* Current instance          */
USERDATA UserData;              /* User structure for I/O operation */

/*---[WinMain]---------------------------------------------------------------
  Syntax:     L_INT L_PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, 
                                     LPSTR lpCmdLine, L_INT nCmdShow) 

  Parameters: hInstance	      Current instance	          
              hPrevInstance 	previous instance	     
              lpCmdLine	 	   command line		     
              nCmdShow 		   show-window type  

  ProtoType:  Decombit.h

  Notes:      Windows main function, calls initialization function 
              and processes message loop
  --------------------------------------------------------------------------*/

L_INT L_PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, 
                       LPSTR lpCmdLine, L_INT nCmdShow) 
 {
 MSG msg;		  /* Message structure, See Windows SDK for more information.*/

 if (!hPrevInstance)			         /* Other instances of app running? */
  if (!InitApplication(hInstance))  /* Initialize shared things        */
	return (FALSE);		            /* Exits if unable to initialize   */

 _fmemcpy (pData.szFilename, lpCmdLine, _fstrlen(lpCmdLine)); /* Copy file name */

 if (!InitInstance(hInstance, nCmdShow)) /* Perform instance initializations */
  return (FALSE);

   /* 
   Acquire and dispatch messages until a WM_QUIT message is received. 
   */
 while (GetMessage(&msg, NULL, NULL, NULL))	
  {
  TranslateMessage(&msg);	   /* Translates virtual key codes	     */
  DispatchMessage(&msg);	   /* Dispatches message to window	     */
  }
 return (msg.wParam);	      /* Returns the value from PostQuitMessage */
 }

/*---[InitApplication]------------------------------------------------------
  Syntax:     L_BOOL InitApplication(HANDLE hInstance)

  Parameters: hInstance	   Current instance	          

  ProtoType:  Decombit.h

  Notes:      Initializes window class structure and registers window class
  --------------------------------------------------------------------------*/

L_BOOL InitApplication(HANDLE hInstance)
 {
 WNDCLASS  wcWindowClass;

 wcWindowClass.style = NULL;                    /* Class style(s).                    */
 wcWindowClass.lpfnWndProc = MainWndProc;       /* Function to retrieve messages for  */
                                                /* windows of this class.             */
 wcWindowClass.cbClsExtra = 0;                  /* No per-class extra data.           */
 wcWindowClass.cbWndExtra = 0;                  /* No per-window extra data.          */
 wcWindowClass.hInstance = hInstance;           /* Application that owns the class.   */
 wcWindowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
 wcWindowClass.hCursor = LoadCursor(NULL, IDC_ARROW);
 wcWindowClass.hbrBackground = GetStockObject(WHITE_BRUSH); 
 wcWindowClass.lpszMenuName =  NULL;            /* No menu */
 wcWindowClass.lpszClassName = "LEADWClass";    /* Name used in call to CreateWindow. */

 return (RegisterClass(&wcWindowClass));        /* Register the window class and return the result code. */
 }

/*---[InitInstance]----------------------------------------------------------
  Syntax:     L_BOOL InitInstance(HANDLE hInstance, L_INT nCmdShow)
                                          
  Parameters: hInstance	   Current instance	          
              nCmdShow     Param for first ShowWindow() call. 

  ProtoType:  Decombit.h

  Notes:      Saves instance handle and creates main window
  --------------------------------------------------------------------------*/

L_BOOL InitInstance(HANDLE hInstance, L_INT nCmdShow)
 {
   /* 
   Save the instance handle in static variable, which will be used in  
   many subsequence calls from this application to Windows.            
   */
 hInst = hInstance;

   /* 
   Create a main window for this application instance.  
   */
 if (L_FileInfo (pData.szFilename, &pData.FileInfo)!=SUCCESS)
   return(FALSE);

 if (!(hWndFrame = CreateWindow(
        "LEADWClass",                     
        "LEAD TOOLS Sample Application",  /* Window title                  */
        WS_OVERLAPPEDWINDOW,              /* Window style.                 */
        CW_USEDEFAULT,                    /* Default horizontal position.  */
        CW_USEDEFAULT,                    /* Default vertical position.    */
        pData.FileInfo.Width,             /* Image width.                  */
        pData.FileInfo.Height,            /* Image height.                 */
        NULL,                             /* Overlapped windows have no parent. */
        NULL,                             /* Use the window class menu.    */
        hInstance,                        /* This instance owns this window.*/
        NULL                              /* Pointer not needed.           */
        )))
  return (FALSE); /* If window could not be created, return "failure" */

 ShowWindow(hWndFrame, nCmdShow);         /* Show the window               */
 UpdateWindow(hWndFrame);                 /* Sends WM_PAINT message        */
 return (TRUE);                           
 }

/*---[MainWndProc]-----------------------------------------------------------
  Syntax:     L_INT32 EXT_FUNCTION MainWndProc(HWND hWnd, L_UINT message, 
                                               WPARAM wParam, LPARAM lParam)
                                          
  Parameters: hWnd	      Window handle		   
              message		Type of message		
              wParam			Additional information	       
              lParam			Additional information	       

  ProtoType:  Decombit.h

  Notes:      This procedure is responsible for handling window messages.
  --------------------------------------------------------------------------*/
L_INT32 EXT_FUNCTION MainWndProc(HWND hWnd, L_UINT Message, WPARAM wParam, 
                                 LPARAM lParam)
 {
 PAINTSTRUCT ps;
 HPALETTE hOldPal = NULL;
 HDC hdc;
 int nRet;
 L_STATIC FILEREADCALLBACK lpfnCallBackRead;
 L_STATIC FILEREADCALLBACK lpfnCallBackOpen;
 L_STATIC FILEREADCALLBACK lpfnCallBackWrite;
 L_STATIC FILEREADCALLBACK lpfnCallBackSeek;
 L_STATIC FILEREADCALLBACK lpfnCallBackClose;

 switch (Message) 
  {
  case WM_PAINT:
   if (pData.bValidBitmap)                /* Do we have an image */
    {
    hdc = BeginPaint (hWnd, &ps);         /* Get DC */
    GetClientRect (hWnd, &pData.BitmapHandle.BitmapRect); /* Get the rectangle to paint */
    if (ACTIVEPALETTE (&pData.BitmapHandle)) /* Is Fixed or optimized palette is active */
     hOldPal = SelectPalette (hdc, ACTIVEPALETTE (&pData.BitmapHandle), TRUE); /* Select it */
    RealizePalette (hdc);  /* Let Windows realize the selected palette */
    L_PaintDevice (hWnd, hdc, &pData.BitmapHandle, &ps.rcPaint, NULL, NULL); /* Paint it */
    if (hOldPal)  /* Return old palette if there is */
     SelectPalette (hdc, hOldPal, FALSE);
    EndPaint (hWnd, &ps);        /* Return DC */
    }
   return (0);

  case WM_CREATE:
   {
      /*
      We will allocate a buffer of the image size and read the image into 
      it, then all our I/O operations will be operate on this buffer 
      */ 
   HFILE hFile = _lopen(pData.szFilename, READ);
   UserData.dwSize = (L_INT32)_llseek(hFile, 0, 2);
   _llseek(hFile, 0, 0);
   UserData.pData = (L_CHAR L_HUGE *)_fmalloc((size_t)UserData.dwSize);
   UserData.pCurData = UserData.pData;
   UserData.dwUsed = 0;
   _lread (hFile, (L_VOID L_HUGE *)UserData.pData, (size_t)UserData.dwSize);
   _lclose(hFile);

      /*
      Create instances of each I/O function.
      */
   lpfnCallBackOpen = (FILEREADCALLBACK) MakeProcInstance ((FARPROC) WindowsOpen, hInst );
   lpfnCallBackRead = (FILEREADCALLBACK) MakeProcInstance ((FARPROC) WindowsRead, hInst );
   lpfnCallBackWrite =(FILEREADCALLBACK) MakeProcInstance ((FARPROC) WindowsWrite, hInst);
   lpfnCallBackSeek = (FILEREADCALLBACK) MakeProcInstance ((FARPROC) WindowsSeek, hInst );
   lpfnCallBackClose =(FILEREADCALLBACK) MakeProcInstance ((FARPROC) WindowsClose, hInst);

      /*
      Initialize the bitmap handle to the image width, height, and 8 
      Bits per pixel 
      */
   L_InitBitmap(&pData.BitmapHandle, pData.FileInfo.Width, pData.FileInfo.Height, 8); 

      /* 
      Use our own routines to replace the internal I/O calls 
      */
   L_RedirectIO((REDIRECTOPEN)lpfnCallBackOpen, (REDIRECTREAD)lpfnCallBackRead, 
                (REDIRECTWRITE)lpfnCallBackWrite, (REDIRECTSEEK)lpfnCallBackSeek,
                (REDIRECTCLOSE)lpfnCallBackClose);
   
      /*
     Decompress the bitmap as an 8 bit
      */
   nRet = L_DecompressBitmap (pData.szFilename, &pData.BitmapHandle, BIT8);

   if (nRet == SUCCESS)
    pData.bValidBitmap = TRUE;
   else
    {
    L_CHAR buf[80];              /* buffer to hold the error message */
    wsprintf( buf, "Error %d Decompressing %s", nRet, pData.szFilename ) ;
	 MessageBox(hWndFrame, buf,"Error", MB_OK ); 
    PostQuitMessage(0);    /* We have an error, so post WM_QUIT, to end the application */
    }
   }
   break;

  case WM_DESTROY:
   FreeProcInstance ((FARPROC) lpfnCallBackOpen);
   FreeProcInstance ((FARPROC) lpfnCallBackRead);
   FreeProcInstance ((FARPROC) lpfnCallBackWrite);
   FreeProcInstance ((FARPROC) lpfnCallBackClose);
   FreeProcInstance ((FARPROC) lpfnCallBackSeek);
   L_FreeBitmap(&pData.BitmapHandle); /* Free the bitmap handle */
	PostQuitMessage(0);                /* Post WM_QUIT, to end the application */
	break;

  default:			  
	return (DefWindowProc(hWnd, Message, wParam, lParam));
  }
 return (NULL);
 }

/*---[WindowsOpen]-----------------------------------------------------------
  Syntax:      L_INT16 L_FAR WindowsOpen(const L_CHAR L_FAR *pFile, 
                                         L_INT16 nMode, L_INT16 nShare)

  Parameters: Check your 'C' manual for more details.

  ProtoType:  Redirect.h 

  Notes:      This procedure is a replacement to the built in Open procedure.
              It will return any number as 5, just to cheat the caller
              that the function is succesfully returned.
  --------------------------------------------------------------------------*/

L_INT16 L_FAR _export WindowsOpen(const L_CHAR L_FAR *pFile, L_INT16 nMode, L_INT16 nShare)
 {
 return(5) ;
 } 

/*---[WindowsRead]-----------------------------------------------------------
  Syntax:      L_UINT16 L_FAR WindowsRead(L_INT16 FD, L_CHAR L_FAR *pBuf, 
                                          L_INT16 nCount)

  Parameters: Check your SDK manual for more details.

  ProtoType:  Redirect.h

  Notes:      This procedure is a replacement to the built in read procedure.
              This will read the data from the allocted memory and adjusting
              the internal data. 
  --------------------------------------------------------------------------*/

L_UINT16 L_FAR _export WindowsRead(L_INT16 FD, L_CHAR L_FAR *pBuf, L_INT16 nCount)
 {
 /* Start reading from the memory */

 if ((L_UINT16)nCount + UserData.pCurData > UserData.pData + UserData.dwSize)
  nCount = (L_INT16)(UserData.dwSize - UserData.dwUsed);
 _fmemcpy (pBuf, UserData.pCurData, (L_UINT16)nCount);
 UserData.pCurData += (L_UINT16)nCount;
 UserData.dwUsed += (L_UINT16)nCount;
 return(nCount);
 }

/*---[WindowsWrite]---------------------------------------------------------
  Syntax:      L_INT16 L_FAR WindowsWrite(L_INT16 FD, L_CHAR L_FAR *pBuf, 
                                          L_INT16 nCount)

  Parameters: Check your SDK manual for more details.

  ProtoType:  Redirect.h

  Notes:      This procedure is a replacement to the built in Write procedure.
              In this example no I/O operation will be done, so we do
              nothing.
  --------------------------------------------------------------------------*/

L_INT16 L_FAR _export WindowsWrite(L_INT16 FD, L_CHAR L_FAR *pBuf, L_INT16 nCount)
 {
 return(nCount);
 }

/*---[WindowsSeek]-----------------------------------------------------------
  Syntax:      L_INT32 L_FAR WindowsSeek(L_INT FD, L_INT32 lnPos, 
                                         L_INT16 nCount)

  Parameters: Check your SDK manual for more details.

  ProtoType:  Redirect.h

  Notes:      This procedure is a replacement to the built-in seek procedure.
              This emulates the whole seek operation on memory block, and 
              adjusting the internal structure according to that. 
  --------------------------------------------------------------------------*/

L_INT32 L_FAR _export WindowsSeek(L_INT FD, L_INT32 lnPos, L_INT16 nOrigin)
 {
 switch (nOrigin)
  {
  case 0:
   UserData.pCurData = UserData.pData + lnPos;
   UserData.dwUsed = lnPos;
   break;
  case 1:
   UserData.pCurData += lnPos;
   UserData.dwUsed += lnPos;
   break;
  case 2:
   assert (0);
   UserData.pCurData = UserData.pData + UserData.dwSize - lnPos;
   UserData.dwUsed = UserData.dwSize - lnPos;
   break;
  }
 return(UserData.pCurData - UserData.pData);
 }

/*---[WindowsClose]-----------------------------------------------------------
  Syntax:      L_INT32 L_FAR WindowsClose(L_INT16 FD)

  Parameters: Check your SDK manual for more details.

  ProtoType:  Redirect.h

  Notes:      This procedure is a replacement to the built in Close procedure.
              This may be called many times especially by L_FileInfo, so 
              we need to reset the internal structure to default. 
  --------------------------------------------------------------------------*/

L_INT16 L_FAR _export WindowsClose(L_INT16 FD)
 {
 UserData.pCurData = UserData.pData;
 UserData.dwUsed = 0;
 return(TRUE);
 }

